/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::ExampleClient
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"

#include <chrono>
#include <thread>

#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <armarx/navigation/client/PathBuilder.h>
#include <armarx/navigation/global_planning/AStar.h>
#include <armarx/navigation/trajectory_control/TrajectoryFollowingController.h>


namespace armarx::navigation::components::example_client
{

    Component::Component()
    {
        // pass
    }


    Component::~Component()
    {
        // pass
    }


    void
    Component::onInitComponent()
    {
        // pass
    }


    void
    Component::onConnectComponent()
    {
        task = new armarx::RunningTask<Component>(this, &Component::exampleNavigationComplex);
        task->start();
    }


    void
    Component::onDisconnectComponent()
    {
        const bool join = true;
        task->stop(join);
    }


    void
    Component::onExitComponent()
    {
        // pass
    }


    std::string
    Component::getDefaultName() const
    {
        return "ExampleClient";
    }


    void
    Component::exampleNavigation()
    {
        using namespace std::chrono_literals;

        // Import relevant namespaces.
        using namespace armarx::navigation;

        ARMARX_INFO << "Configuring navigator";

        // Create an example configuration valid for the following move* calls.
        configureNavigator(
            client::NavigationStackConfig()
                .general({} /*{.maxVel = VelocityLimits{.linear = 400 , .angular = 0.1}}*/)
                .globalPlanner(glob_plan::AStarParams())
                .trajectoryController(traj_ctrl::TrajectoryFollowingControllerParams()));

        // Example of registering a lambda as callback.
        navigator.onGoalReached([&]() { ARMARX_IMPORTANT << "Goal reached! (lambda-style)"; });

        // Example of registering a method as callback.
        navigator.onGoalReached([this] { goalReached(); });

        std::this_thread::sleep_for(1s);

        ARMARX_INFO << "Moving to goal pose";
        // Start moving to goal position using above config.
        core::Pose goal = core::Pose::Identity();
        goal.translation() << 2000, 1000, 0;

        navigator.moveTo(goal, core::NavigationFrame::Absolute);

        std::this_thread::sleep_for(15s);

        // Wait until goal is reached
        armarx::navigation::client::StopEvent se = navigator.waitForStop();
        if (se)
        {
            ARMARX_INFO << "Goal 1 reached.";
        }
        else
        {
            if (se.isSafetyStopTriggeredEvent())
            {
                ARMARX_ERROR << "Safety stop was triggered!";
            }
            else if (se.isUserAbortTriggeredEvent())
            {
                ARMARX_ERROR << "Aborted by user!";
            }
            else if (se.isInternalErrorEvent())
            {
                ARMARX_ERROR << "Unknown internal error occured! "
                             << se.toInternalErrorEvent().message;
            }
        }

        goal.translation() << -1500, 1000, 0;
        navigator.moveTo(goal, core::NavigationFrame::Absolute);

        std::this_thread::sleep_for(15s);

        // Wait until goal is reached
        se = navigator.waitForStop();
        if (se)
        {
            ARMARX_INFO << "Goal 2 reached.";
        }
        else
        {
            ARMARX_ERROR << "Could not reach goal 2!";
        }

        goal.translation() << 4500, 4500, 0;
        navigator.moveTo(goal, core::NavigationFrame::Absolute);

        // Wait until goal is reached
        se = navigator.waitForStop();
        if (se)
        {
            ARMARX_INFO << "Goal 3 reached.";
        }
        else
        {
            ARMARX_ERROR << "Could not reach goal 3!";
        }

        std::this_thread::sleep_for(15s);

        // TODO example with waypoints

        ARMARX_INFO << "Moving into certain direction.";
        // Start moving towards a direction
        navigator.moveTowards(core::Direction(100, 100, 0), core::NavigationFrame::Relative);

        std::this_thread::sleep_for(3s);

        ARMARX_INFO << "Pausing movement.";
        navigator.pause();
    }

    void
    Component::exampleNavigationComplex()
    {
        ARMARX_INFO << "Demonstrating complex navigation";

        using namespace std::chrono_literals;

        // Import relevant namespaces.
        using namespace armarx::navigation;

        ARMARX_INFO << "Configuring navigator";

        // Create an example configuration valid for the following move* calls.
        configureNavigator(
            client::NavigationStackConfig()
                .general({} /*{.maxVel = VelocityLimits{.linear = 400 , .angular = 0.1}}*/)
                .globalPlanner(glob_plan::AStarParams())
                .trajectoryController(traj_ctrl::TrajectoryFollowingControllerParams()));

        // Example of registering a lambda as callback.
        navigator.onGoalReached([&]() { ARMARX_IMPORTANT << "Goal reached! (lambda-style)"; });

        // Example of registering a method as callback.
        navigator.onGoalReached([this] { goalReached(); });

        std::this_thread::sleep_for(1s);

        ARMARX_INFO << "Moving to goal pose";
        // Start moving to goal position using above config.

        core::Pose anonymousWaypoint = core::Pose::Identity();
        anonymousWaypoint.translation() << 2000, 1000, 0;

        core::Pose goal = core::Pose::Identity();
        goal.translation() << 2000, 1500, 0;

        client::PathBuilder pathBuilder;

        // clang-format off
        pathBuilder
            .add(anonymousWaypoint)
            .add("Testing/home")
            .add(goal);
        // clang-format on

        ARMARX_INFO << "Starting execution";
        navigator.moveTo(pathBuilder, core::NavigationFrame::Absolute);

        std::this_thread::sleep_for(25s); // TODO(fabian.reister): remove in the future

        // Wait until goal is reached
        armarx::navigation::client::StopEvent se = navigator.waitForStop();
        if (se)
        {
            ARMARX_INFO << "Goal 1 reached.";
        }
        else
        {
            if (se.isSafetyStopTriggeredEvent())
            {
                ARMARX_ERROR << "Safety stop was triggered!";
            }
            else if (se.isUserAbortTriggeredEvent())
            {
                ARMARX_ERROR << "Aborted by user!";
            }
            else if (se.isInternalErrorEvent())
            {
                ARMARX_ERROR << "Unknown internal error occured! "
                             << se.toInternalErrorEvent().message;
            }
        }

        ARMARX_INFO << "Pausing movement.";
        navigator.pause();
    }


    void
    Component::goalReached()
    {
        ARMARX_IMPORTANT << "Goal reached! (method-style)";
    }


    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new armarx::ComponentPropertyDefinitions(getConfigIdentifier());
        return def;
    }


    ARMARX_DECOUPLED_REGISTER_COMPONENT(Component);

} // namespace armarx::navigation::components::example_client
